"""
Defines typing for structures having specific meaning.
"""

from __future__ import annotations

from typing import Union

import pygame

Position = Union[pygame.Vector2, tuple[int, int]]
"""Alias for pygame.Vector2. A tuple[int, int] can alternatively be given in place of it."""
Margin = tuple[int, int, int, int]
"""Alias for tuple[int, int, int, int]"""



"""
Defines InfoBox class, an helper to draw any kind of popup (menu or informative message).
"""

from __future__ import annotations

import os.path
from typing import Union, Sequence, Callable, Optional

import pygame

from ..configuration import _default_sprites, _default_fonts, _default_texts
from ..constants import (
    WHITE,
    CLOSE_BUTTON_MARGIN_TOP,
    MARGIN_BOX,
    DEFAULT_MARGIN_TOP,
    CLOSE_BUTTON_SIZE,
    MARGIN_LINKED_ELEMENT,
    DEFAULT_POPUP_WIDTH,
)
from .box_element import BoxElement
from .text_element import TextElement
from .button import Button
from ..types import Position


class _Row:
    def __init__(self, elements: list[BoxElement], height: int = 0):
        self.elements = elements
        self.height = height


class InfoBox:
    """
    This class is defining any kind of popup that can be found in the app.

    It can be used to represent the interface of a menu, or a simple text message.
    Some elements can be buttons, that will react to user clicks (see the button component
    for more information).


    Keyword arguments:
        title (str): the title of the infoBox
        element_grid (list[list[BoxElement]]): a grid containing the components that should be rendered by the infoBox
        width (int): the width of the infoBox, defaults to DEFAULT_POPUP_WIDTH
        element_linked (pygame.Rect): the pygame Rect of the element linked to this infoBox,
            the infoBox will be displayed beside the element if provided
        position (Position): the static position of the infoBox, if not provided position would be computed basing on
        element linked or screen
        has_close_button (bool): whether a close button should be added at the bottom or not, defaults to True
        title_color (pygame.Color): the color of the title
        background_path (str): the path corresponding to the image that should be the sprite of the infoBox
        close_button_text (str): the text that will be shown on close button
        close_button_background_path (str): the path to the image corresponding to the sprite of the close button
            if there should be one
        close_button_background_hover_path (str): the path to the image corresponding to the sprite of
            the close button when it is hovered if there should be one
        visible_on_background (bool): whether the popup is visible on background or not, defaults to True
        has_vertical_separator (bool): whether there should be a line splitting the infoBox in two at middle width or
            not, defaults to False
        identifier (str): a string permitting to identify the menu among others if needed

    Attributes:
        title (str): the title of the infoBox
        element_linked (pygame.Rect): the pygame Rect of the element linked to this infoBox if there is one
        has_close_button (bool): whether the infoBox has a close button or not
        title_color (pygame.Color): the color of the title
        element_grid (list[list[BoxElement]): the grid containing the components that should be rendered by the infoBox
        buttons (Sequence[Button]): the sequence of buttons of the infoBox, including the close button if present
        sprite (pygame.Surface): the pygame Surface corresponding to the sprite of the infoBox
        visible_on_background (bool): whether the popup is visible on background or not
        identifier (str): a string permitting to identify the menu among others if needed
    """

    def __init__(
        self,
        title: str,
        element_grid: list[list[BoxElement]],
        width: int = DEFAULT_POPUP_WIDTH,
        element_linked: Optional[pygame.Rect] = None,
        position: Optional[Position] = None,
        has_close_button: bool = True,
        title_color: pygame.Color = WHITE,
        background_path: str = None,
        close_button_text: Optional[str] = None,
        close_button_background_path: str = None,
        close_button_background_hover_path: str = None,
        visible_on_background: bool = True,
        has_vertical_separator: bool = False,
        identifier: str = "",
    ) -> None:
        self.title: str = title
        self.element_linked: pygame.Rect = element_linked
        self.has_close_button: bool = has_close_button
        self.title_color: pygame.Color = title_color
        self.element_grid: list[list[BoxElement]] = element_grid
        self.__separator: dict[str, Union[bool, int]] = {
            "display": has_vertical_separator,
            "vertical_position": 0,
            "height": 0,
        }
        background_path = (
            os.path.abspath(background_path)
            if background_path
            else _default_sprites["info_box_background"]
        )
        self.sprite: pygame.Surface = pygame.image.load(background_path)
        self.close_button_text: str = close_button_text if close_button_text is not None \
            else _default_texts["close_button"]
        self.__close_button_background_path: str = close_button_background_path
        self.__close_button_background_hover_path: str = (
            close_button_background_hover_path
        )
        self.__elements: list[_Row] = self.init_elements()
        self.buttons: Sequence[Button] = []
        self.__size: tuple[int, int] = (width, 0)
        self.__is_position_static: bool = position is not None
        if self.__is_position_static:
            self.position: Optional[Position] = pygame.Vector2(position)
        else:
            self.position = None
        self.visible_on_background: bool = visible_on_background
        self.identifier: str = identifier

    def __repr__(self):
        return f"InfoBox with identifier '{self.identifier}'"

    def init_render(
        self, screen: pygame.Surface, close_button_callback: Callable = None
    ) -> None:
        """
        Initialize the rendering of the popup.

        Compute it size and its position according to the given screen.
        Determine the position of each component.

        Keyword arguments:
            screen (pygame.Surface): the screen on which the popup is
            close_button_callback (Callable): the callback that should be executed when clicking on
                the close button if there is any
        """
        if self.has_close_button:
            self.__elements[-1].elements[0].callback = close_button_callback
        self.__resize_elements()
        height: int = self.__determine_height()
        self.__size = (self.__size[0], height)
        if not self.__is_position_static:
            self.position = self.determine_position(screen)
        if self.position is not None:
            self.determine_elements_position()
        self.buttons = self.find_buttons()
        self.sprite = pygame.transform.scale(self.sprite.convert_alpha(), self.__size)
        self.__separator["height"] += height

    def init_elements(self) -> list[_Row]:
        """
        Initialize the graphical elements associated to the formal data that the infoBox should
        represent.

        Returns:
             the elements in a 2D structure corresponding to the relative position of each element.
        """
        elements: list[_Row] = [
            _Row(element_line) for element_line in self.element_grid
        ]
        title = TextElement(
            self.title,
            pygame.Vector2(0, 0),
            _default_fonts["info_box_title"],
            (20, 0, 20, 0),
            self.title_color,
        )
        self.__separator["vertical_position"] += title.get_height()
        elements.insert(0, _Row([title]))
        if self.has_close_button:
            elements.append(
                _Row(
                    [
                        Button(
                            size=CLOSE_BUTTON_SIZE,
                            title=self.close_button_text,
                            background_path=self.__close_button_background_path,
                            background_hover_path=self.__close_button_background_hover_path,
                            margin=(CLOSE_BUTTON_MARGIN_TOP, 0, 0, 0),
                        )
                    ]
                )
            )
        return elements

    def __determine_height(self) -> int:
        """
        Compute the total height of the infoBox, defined according
        to the height of each element in it.

        Returns:
             int: the computed height.
        """
        # Margin to be add at begin and at end
        height: int = MARGIN_BOX * 2
        self.__separator["height"] -= height
        self.__separator["vertical_position"] += height

        for row in self.__elements:
            max_height: int = 0
            for element in row.elements:
                el_height = element.get_height() + DEFAULT_MARGIN_TOP
                if el_height > max_height:
                    max_height = el_height
            height += max_height
            row.height = max_height

        if self.has_close_button:
            self.__separator["height"] -= self.__elements[-1].height

        return height

    def __resize_elements(self) -> None:
        """
        Resize elements according to the current width of the infoBox
        """
        for row in self.__elements:
            for element in row.elements:
                if isinstance(element, TextElement):
                    element.content = element._verify_rendered_text_size(
                        element.content,
                        element._text,
                        self.__size[0]
                        - element.get_margin_left()
                        - element.get_margin_right(),
                    )
                    element.size = element.content.get_size()

    def determine_position(self, screen: pygame.Surface) -> Optional[Position]:
        """
        Compute the position of the infoBox to be beside the linked element.

        If no element is linked to the infoBox, the position will be determined at display time
        according to the screen.

        Returns:
             Optional[Position]: the computed position.

        Keyword arguments:
            screen (pygame.Surface): The screen on which the infoBox is rendered.
        """
        if self.element_linked:
            position: Position = pygame.Vector2(
                self.element_linked.x
                + self.element_linked.width
                + MARGIN_LINKED_ELEMENT,
                self.element_linked.y
                + self.element_linked.height // 2
                - self.__size[1] // 2,
            )
            if position.y < 0:
                position.y = 0
            elif position.y + self.__size[1] > screen.get_height():
                position.y = screen.get_height() - self.__size[1]
            if position.x + self.__size[0] > screen.get_width():
                position.x = self.element_linked.x - self.__size[0]
            return position
        return None

    def find_buttons(self) -> Sequence[Button]:
        """
        Search in all elements for buttons.

        Returns:
             Sequence[Button]: the sequence of buttons.
        """
        buttons: list[Button] = []
        for row in self.__elements:
            for element in row.elements:
                if isinstance(element, Button):
                    buttons.append(element)
        return buttons

    def determine_elements_position(self) -> None:
        """
        Compute the position of each element and update it if needed.
        """
        y_coordinate: int = self.position[1] + MARGIN_BOX
        # Memorize mouse position in case it is over a button
        mouse_pos = pygame.mouse.get_pos()
        # A row begins by a value identifying its height, followed by its elements
        for row in self.__elements:
            nb_elements = len(row.elements)
            i = 1
            for element in row.elements:
                base_x = self.position.x + (self.__size[0] // (2 * nb_elements)) * i
                x_coordinate = base_x - element.get_width() // 2
                element.position = pygame.Vector2(
                    x_coordinate,
                    y_coordinate + element.get_margin_top(),
                )
                if isinstance(element, Button):
                    element.set_hover(element.get_rect().collidepoint(mouse_pos))
                i += 2
            y_coordinate += row.height

    def display(self, screen: pygame.Surface) -> None:
        """
        Display the infoBox and all its elements.

        Keyword arguments:
            screen (pygame.Surface): the screen on which the displaying should be done
        """
        if self.position is not None:
            screen.blit(self.sprite, self.position)
        else:
            win_size = screen.get_size()
            self.position = pygame.Vector2(
                win_size[0] // 2 - self.__size[0] // 2,
                win_size[1] // 2 - self.__size[1] // 2,
            )
            screen.blit(self.sprite, self.position)
            self.determine_elements_position()

        for row in self.__elements:
            for element in row.elements:
                element.display(screen)

        if self.__separator["display"]:
            pygame.draw.line(
                screen,
                WHITE,
                (
                    self.position.x + self.__size[0] / 2,
                    self.position.y + self.__separator["vertical_position"],
                ),
                (
                    self.position.x + self.__size[0] / 2,
                    self.position.y + self.__separator["height"],
                ),
                2,
            )

    def is_position_inside(self, position: Position) -> bool:
        """
        Check if given position is inside or outside the InfoBox.

        Returns:
            bool: whether the position is inside the borders or not

        Keyword arguments:
            position (Position): the position to be checked
        """
        return pygame.Rect(self.position, self.__size).collidepoint(position)

    def click(self, position: Position) -> Callable:
        """
        Handle the triggering of a click event.

        Returns:
            Optional[Callable]: the data corresponding to the action that should be done if the click was done
                on a button, else None.

        Keyword arguments:
            position (Position): the position of the mouse
        """
        for button in self.buttons:
            if button.get_rect().collidepoint(position):
                return button.action_triggered()
        # Return a "do nothing" callable when clicking on empty space
        return lambda: None

    def motion(self, position: Position) -> None:
        """
        Handle the triggering of a motion event.
        Test if the mouse entered a button or left one.

        Keyword arguments:
            position (Position): the position of the mouse
        """
        for button in self.buttons:
            mouse_is_on_button: bool = button.get_rect().collidepoint(position)
            button.set_hover(mouse_is_on_button and not button.disabled)



from .box_element import BoxElement
from .button import Button
from .dynamic_button import DynamicButton
from .image_button import ImageButton
from .info_box import InfoBox
from .text_element import TextElement



"""
Defines the initialization process of the package through the function init.
"""
from warnings import warn

import pygame

from . import fonts
from ._exceptions.wrong_initialization_exception import WrongInitializationException

_is_initialized = False


def init() -> None:
    """
    Initialize the package modules.

    Must be called before all other calls and after the initialization of pygame.
    """
    global _is_initialized
    if _is_initialized:
        warn("pygamepopup is already initialized, no need to initialize it again")
        return

    if not pygame.get_init():
        raise WrongInitializationException(
            "pygamepopup.init() has to be called only after pygame.init()"
        )

    fonts._init()
    _is_initialized = True



"""
Defines BoxElement class, useful for drawing elements.
All other components that should be drawn on a popup inherit from this class.
"""

from __future__ import annotations

from typing import Optional

import pygame

from .. import initialization
from .._exceptions.wrong_initialization_exception import WrongInitializationException
from ..types import Position, Margin


class BoxElement:
    """
    This element acts as a wrapper for a gui element.
    In fact, it adds a margin to any border of an element to have a cleaner interface.

    Keyword arguments:
        position (Position): the position of the box on the screen
        content (Optional[pygame.Surface]): a surface that will be wrapped by the box
        margin (Margin): a tuple containing the margins of the box, should be in the form
            "(top_margin, right_margin, bottom_margin, left_margin), defaults to (0, 0, 0, 0)"

    Attributes:
        position (Position): the position of the box on the screen
        content (Optional[pygame.Surface]): the element wrapped in the box
        size (tuple[int, int]): the size of the content following the format "(width, height)"
        margin (dict[str, int]): a dict containing all the values for margins TOP, BOTTOM, LEFT and RIGHT
    """

    def __new__(cls, *args, **kwargs):
        if not initialization._is_initialized:
            raise WrongInitializationException(
                "pygamepopup.init() has to be called before any other interaction with "
                "pygamepopup"
            )
        return super().__new__(cls)

    def __init__(
        self,
        position: Position,
        content: Optional[pygame.Surface],
        margin: Margin = (0, 0, 0, 0),
    ) -> None:
        self.position: Position = position
        self.content: pygame.Surface = content
        self.size: tuple[int, int] = (0, 0)
        if self.content:
            self.size = (self.content.get_width(), self.content.get_height())
        self.margin: dict[str, int] = {
            "TOP": margin[0],
            "BOTTOM": margin[2],
            "LEFT": margin[3],
            "RIGHT": margin[1],
        }

    def get_width(self) -> int:
        """
        Returns:
             int: the width of the content more the left and right margins
        """
        return self.margin["LEFT"] + self.size[0] + self.margin["RIGHT"]

    def get_height(self) -> int:
        """
        Returns:
             int: the height of the content more the top and bottom margins
        """
        return self.margin["TOP"] + self.size[1] + self.margin["BOTTOM"]

    def get_margin_top(self) -> int:
        """
        Returns:
             int: top margin
        """
        return self.margin["TOP"]

    def get_margin_bottom(self) -> int:
        """
        Returns:
             int: bottom margin
        """
        return self.margin["BOTTOM"]

    def get_margin_left(self) -> int:
        """
        Returns:
             int: left margin
        """
        return self.margin["LEFT"]

    def get_margin_right(self) -> int:
        """
        Returns:
             int: right margin
        """
        return self.margin["RIGHT"]

    def get_rect(self) -> pygame.Rect:
        """
        Returns:
             pygame.Rect: a pygame rect containing the position of the element and its size
        """
        return pygame.Rect(
            self.position[0] + self.margin["LEFT"],
            self.position[1] + self.margin["TOP"],
            self.size[0],
            self.size[1],
        )

    def display(self, screen: pygame.Surface) -> None:
        """
        Display the content of the box, following the margins that should be added around it.

        Keyword arguments:
            screen (pygame.Surface): the screen on which the content of the box should be drawn
        """
        screen.blit(
            self.content,
            (
                self.position[0] + self.margin["LEFT"],
                self.position[1] + self.margin["TOP"],
            ),
        )



"""
Defines MenuManager class, the main class to handle the displaying of all menus on a screen and
handle the triggering of user events on the elements of the active menu
"""

from __future__ import annotations

from typing import Optional, Sequence

import pygame

from .components.info_box import InfoBox
from .types import Position


class MenuManager:
    """
    This class represents a manager for the interfaces of a screen.

    It displays all the menus currently registered in the manager.

    Handle the opening of a new menu and the closing of the active one.
    Handle the triggering of user motion events and user click events on the active menu.

    Keyword arguments:
        screen (pygame.Surface): the screen on which the menus should be displayed and on which the
            user events should be handled

    Attributes:
        screen (pygame.Surface): the screen on which the menus should be displayed and on which the
            user events should be handled
        active_menu (Optional[InfoBox]): the current menu in the foreground, the only one that will react to user events
        background_menus (list[InfoBox]): the ordered sequence of menus that are in the background
    """

    def __init__(self, screen: pygame.Surface) -> None:
        self.screen: pygame.Surface = screen
        self.active_menu: Optional[InfoBox] = None
        self.background_menus: list[InfoBox] = []

    def open_menu(self, menu: InfoBox) -> None:
        """
        Open the given menu.
        Initialize the rendering of the menu, and set it as the new active menu.

        Previous active menu is sent to the background.

        Keyword arguments:
            menu (InfoBox): the popup that should be open
        """
        self._prepare_menu(menu)
        if self.active_menu:
            self.background_menus.append(self.active_menu)
        self.active_menu = menu

    def replace_given_menu(
        self, menu_identifier: str, new_menu: InfoBox, all_occurrences: bool = False
    ) -> bool:
        """
        Replace a menu by a new one according to its identifier.

        By default, only first occurrence is replaced if many menus in the manager have the given identifier.

        Keyword arguments:
            menu_identifier (str): the identifier of the menu to be replaced
            new_menu (InfoBox): the menu that should replace the given one
            all_occurrences (bool): whether all found occurrences should be replaced or only the first one

        Returns:
             bool: whether the replacement has succeeded or not
        """
        has_replacement_been_done = False
        if self.active_menu and self.active_menu.identifier == menu_identifier:
            self._prepare_menu(new_menu)
            self.active_menu = new_menu
            if not all_occurrences:
                return True
            has_replacement_been_done = True
        for index, menu in enumerate(self.background_menus):
            if menu.identifier == menu_identifier:
                self._prepare_menu(new_menu)
                self.background_menus[index] = new_menu
                if not all_occurrences:
                    return True
                has_replacement_been_done = True
        return has_replacement_been_done

    def close_active_menu(self) -> None:
        """
        Close the active menu by 'destroying' it.

        Take the next menu in the background to move it to foreground if there is any.
        """
        self.active_menu = (
            self.background_menus.pop() if len(self.background_menus) != 0 else None
        )
        if self.active_menu:
            # Trigger an irrelevant motion event to refresh the hovering of buttons on the new menu
            self.active_menu.motion(pygame.Vector2(pygame.mouse.get_pos()))

    def close_given_menu(
        self, menu_identifier: str, all_occurrences: bool = False
    ) -> bool:
        """
        Close menu corresponding to the given identifier.

        By default, only first occurrence is closed if many menus in the manager have the given identifier.

        Keyword arguments:
            menu_identifier (str): the identifier of the menu to be closed
            all_occurrences (bool): whether all found occurrences should be closed or only the first one,
                defaults to False

        Returns:
             bool: whether at least one menu has been closed or not
        """
        if self.active_menu and self.active_menu.identifier == menu_identifier:
            self.active_menu = None
            if not all_occurrences:
                return True

        matching_menus_in_background = self._get_given_menus_from_background(
            menu_identifier
        )
        for menu in matching_menus_in_background:
            self.background_menus.remove(menu)
            if not all_occurrences:
                break
        return len(matching_menus_in_background) > 0

    def clear_menus(self) -> None:
        """
        Close all the menus (in foreground and in background)
        """
        self.active_menu = None
        self.background_menus.clear()

    def reduce_active_menu(self) -> None:
        """
        Move the active menu to the background.
        """
        if self.active_menu:
            self.background_menus.append(self.active_menu)
            self.active_menu = None

    def display(self) -> None:
        """
        Display all the visible menus in the background in order first, then display the active menu
        """
        for menu in self.background_menus:
            if menu.visible_on_background:
                menu.display(self.screen)
        if self.active_menu:
            self.active_menu.display(self.screen)

    def click(self, button: int, position: Position) -> None:
        """
        Handle the triggering of a click event.
        Delegate this event to the active menu if there is any and if it's a left click.

        Keyword arguments:
            button (int): a value representing which mouse button has been pressed
                (1 for left button, 2 for middle button, 3 for right button)
            position (Position): the position of the mouse
        """
        if button == 1:
            if self.active_menu:
                self.active_menu.click(position)()

    def motion(self, position: Position) -> None:
        """
        Handle the triggering of a motion event.
        Delegate this event to the active menu if there is any.

        Keyword arguments:
            position (Position): the position of the mouse
        """
        if self.active_menu:
            self.active_menu.motion(position)

    def _prepare_menu(self, menu: InfoBox) -> None:
        """
        Prepare the given menu to be rendered according to the current setup.

        Keyword arguments:
            menu (InfoBox): the menu to be initialized
        """
        menu.init_render(self.screen, close_button_callback=self.close_active_menu)

    def _get_given_menus_from_background(
        self, menu_identifier: str
    ) -> Sequence[InfoBox]:
        """
        Returns:
             Sequence[InfoBox]: all the menus in background matching the given identifier

        Keyword arguments:
            menu_identifier (str): the identifier to look for
        """
        return [
            menu for menu in self.background_menus if menu.identifier == menu_identifier
        ]



"""
Defines Button class, a BoxElement able to react to user actions.
"""

from __future__ import annotations

import os.path
from enum import Enum
from typing import Union, Callable, Sequence

import pygame

from ..configuration import _default_sprites, _default_fonts
from .box_element import BoxElement
from ..constants import WHITE, BUTTON_SIZE
from ..types import Position, Margin


class Button(BoxElement):
    """
    This class is representing any kind of button that could be seen on an interface.

    A button is receptive to user clicks and returns an id corresponding to a method,
    it may have specific arguments too.

    Mouse motion is also handled: the button appearance can change according to current focus.

    Keyword arguments:
        callback (Callable): the reference to the function that should be call after a click.
        size (tuple[int, int]): the size of the button following the format "(width, height)", defaults to BUTTON_SIZE.
        title (str): the text that should be displayed at the center of the element.
        position (Position): the position of the element on the screen.
        background_path (str): the path to the image corresponding to the sprite of the element.
        background_hover_path (str): the path to the image corresponding to the sprite of the element
            when it has the focus.
        no_background (bool): specify whether a background should be present or not, defaults to False.
        margin (Margin): a tuple containing the margins of the box,
            should be in the form "(top_margin, right_margin, bottom_margin, left_margin)", defaults to (0, 0, 0, 0).
        disabled (bool): a boolean indicating if it is not possible to interact with the button, defaults to False.
        font (pygame.font.Font): the font that should be used to render the text content.
        text_color (pygame.Color): the color of the text content, defautls to WHITE.
        font_hover (pygame.font.Font): the font that should be used to render the text content when the mouse is over
            the button.
        text_hover_color (pygame.Color): the color of the text content when the mouse is over the button,
            defaults to WHITE.
        complementary_text_lines (str): the other text lines that should be displayed in addition of
            the title.

    Attributes:
        callback (Callable): the reference to the function that should be call after a click.
        sprite (pygame.Surface): the pygame Surface corresponding to the sprite of the element.
        sprite_hover (pygame.Surface): the pygame Surface corresponding to the sprite of the element
            when it has the focus.
    """

    def __init__(
        self,
        callback: Callable = lambda: None,
        size: tuple[int, int] = BUTTON_SIZE,
        title: str = "",
        position: Position = pygame.Vector2(0, 0),
        background_path: str = None,
        background_hover_path: str = None,
        no_background: bool = False,
        margin: Margin = (0, 0, 0, 0),
        disabled: bool = False,
        font: pygame.font.Font = None,
        text_color: pygame.Color = WHITE,
        font_hover: pygame.font.Font = None,
        text_hover_color: pygame.Color = WHITE,
        complementary_text_lines: Sequence[str] = None,
    ) -> None:
        super().__init__(position, None, margin)
        self.callback: Union[Enum, Callable] = callback
        self.size: tuple[int, int] = size

        if complementary_text_lines is None:
            complementary_text_lines = []
        text_lines = [title] + complementary_text_lines

        if not font:
            font = _default_fonts["button_title"]
        rendered_text_lines = Button.render_text_lines(text_lines, text_color, font)
        if no_background:
            self.size = (
                rendered_text_lines[0].get_width(),
                rendered_text_lines[0].get_height() * len(rendered_text_lines),
            )

        if no_background:
            background_path = None
        elif background_path:
            background_path = os.path.abspath(background_path)
        else:
            background_path = _default_sprites["button_background"]["inactive"]
        self.sprite = self.render_sprite(background_path, rendered_text_lines)

        if not font_hover:
            font_hover = font
        rendered_text_lines_hover = Button.render_text_lines(
            text_lines, text_hover_color, font_hover
        )

        if no_background:
            background_hover_path = None
        elif background_hover_path:
            background_hover_path = os.path.abspath(background_hover_path)
        else:
            background_hover_path = _default_sprites["button_background"]["active"]
        self.sprite_hover = self.render_sprite(
            background_hover_path, rendered_text_lines_hover
        )

        self.content = self.sprite
        self.disabled = disabled

    @staticmethod
    def render_text_lines(
        text_lines: Sequence[str],
        text_color: pygame.Color,
        font: pygame.font.Font,
    ) -> Sequence[pygame.Surface]:
        """
        Compute the rendering of the given text.

        Returns:
            Sequence[pygame.Surface]: the rendered text lines.

        Keyword arguments:
            text_lines (Sequence[str]): the sequence in order of text lines to be rendered.
            text_color (pygame.Color): the color of the text.
            font (pygame.font.Font): the font that should be used to render the text.
        """
        return [font.render(text_line, True, text_color) for text_line in text_lines]

    def render_sprite(
        self, background_path: str, rendered_text_lines: Sequence[pygame.Surface]
    ) -> pygame.Surface:
        """
        Compute the rendering of the button with the given background and text lines.
        If no background is provided, render the text on an empty surface.

        Returns:
             pygame.Surface: the generated surface.

        Keyword arguments:
            background_path (str): the path to the image corresponding to the sprite of the button.
            rendered_text_lines (Sequence[pygame.Surface]): the sequence of text lines in order that should be clipped
                on the surface.
        """
        raw_sprite = (
            pygame.image.load(background_path)
            if background_path
            else pygame.Surface((0, 0))
        )
        sprite = pygame.transform.scale(raw_sprite.convert_alpha(), self.size)
        text_lines_count = len(rendered_text_lines)

        for index, rendered_text_line in enumerate(rendered_text_lines):
            sprite.blit(
                rendered_text_line,
                (
                    sprite.get_width() // 2 - rendered_text_line.get_width() // 2,
                    (2 * index + 1) * sprite.get_height() // (2 * text_lines_count)
                    - rendered_text_line.get_height() // 2,
                ),
            )
        return sprite

    def set_hover(self, is_mouse_hover: bool) -> None:
        """
        Change the current sprite between sprite or sprite_hover
        depending on whether the mouse is over the element or not.

        Keyword arguments:
            is_mouse_hover (bool): a boolean value indicating if the mouse is over the element or not
        """
        self.content = self.sprite_hover if is_mouse_hover else self.sprite

    def action_triggered(self) -> Callable:
        """
        Method that should be called after a click.

        Returns:
             Callable: the callback that should be executed, a callback doing nothing would be returned
             if the button is disabled.
        """
        if self.disabled:
            return lambda: None
        return self.callback



"""
Defines TextElement class, a BoxElement permitting to easily draw text and let it
be centered on an popup.
"""

from __future__ import annotations

import pygame
from pygame.constants import SRCALPHA

from ..configuration import _default_fonts
from ..constants import WHITE
from .box_element import BoxElement
from ..types import Position, Margin


class TextElement(BoxElement):
    """
    This class is representing a paragraph of text, displayed on a popup and horizontally
    centered according to its position.

    Keyword arguments:
        text (str): the text that should be rendered.
        position (Position): the position of the text on the screen.
        font (pygame.font.Font): the font that should be used to render the text.
        margin (Margin): a tuple containing the margins of the box,
            should be in the form "(top_margin, right_margin, bottom_margin, left_margin)", defaults to (0, 0, 0, 0).
        text_color (pygame.Color): the color of the rendered text, defaults to WHITE.
    """

    def __init__(
        self,
        text: str,
        position: Position = pygame.Vector2(0, 0),
        font: pygame.font.Font = None,
        margin: Margin = (0, 0, 0, 0),
        text_color: pygame.Color = WHITE,
    ) -> None:
        if not font:
            font = _default_fonts["text_element_content"]
        self._font = font
        self._text = text
        self._text_color = text_color
        rendered_text: pygame.Surface = font.render(text, True, text_color)
        super().__init__(position, rendered_text, margin)

    def _verify_rendered_text_size(
        self, rendered_text: pygame.Surface, text: str, container_width: int
    ) -> pygame.Surface:
        """
        Split a given text in multiple lines until it could fit properly in its container.

        Returns:
             pygame.Surface: the final rendered text

        Keyword arguments:
            rendered_text (pygame.Surface): the current rendering of the text, to check if it fits in the container.
            text (str): the text that would be split if necessary.
            container_width (int): the width of the container.
        """
        final_render = rendered_text

        if final_render.get_width() + 20 > container_width:
            first_part, second_part = TextElement.__divide_text(text)
            first_part_render = self._font.render(first_part, True, self._text_color)
            first_part_render = self._verify_rendered_text_size(
                first_part_render, first_part, container_width
            )
            second_part_render = self._font.render(second_part, True, self._text_color)
            second_part_render = self._verify_rendered_text_size(
                second_part_render, second_part, container_width
            )
            final_render = pygame.Surface(
                (
                    container_width,
                    first_part_render.get_height() + second_part_render.get_height(),
                ),
                SRCALPHA,
            )
            first_part_x = (
                final_render.get_width() // 2 - first_part_render.get_width() // 2
            )
            final_render.blit(first_part_render, (first_part_x, 0))
            second_part_x = (
                final_render.get_width() // 2 - second_part_render.get_width() // 2
            )
            final_render.blit(
                second_part_render, (second_part_x, first_part_render.get_height())
            )
        return final_render

    @staticmethod
    def __divide_text(text: str) -> tuple[str, str]:
        """
        Divide a text in two parts of a similar size, avoiding to cut a word in two.

        Returns:
             tuple[str, str]: the split text in two different strings.

        Keyword argument:
            text (str): the text that should be divided.
        """
        separation_index: int = TextElement.__get_middle_text(text)
        return text[:separation_index], text[separation_index:]

    @staticmethod
    def __get_middle_text(text: str) -> int:
        """
        Returns:
             int: the index of the first whitespace character that is after the middle of the provided
            string. If there is no whitespace character after the middle of the string, -1 is returned.

        Keyword attributes:
            text (str): the string that should be used to make the computation.
        """
        absolute_middle = len(text) // 2
        for i in range(absolute_middle, len(text)):
            if text[i] == " ":
                return i
        return -1
